from opendr.perception.semantic_segmentation.bisenet.bisenet_learner import BisenetLearner
from opendr.perception.semantic_segmentation.bisenet.CamVid import CamVidDataset

__all__ = ['BisenetLearner', 'CamVidDataset']



# Copyright 2020-2022 OpenDR European Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from torch.utils.data import DataLoader
from torchvision import transforms
import os
import json
import tqdm
import numpy as np
from imgaug import augmenters as iaa
from PIL import Image as PILImage
from urllib.request import urlretrieve
from opendr.perception.semantic_segmentation.bisenet.algorithm.model.build_BiSeNet import BiSeNet
from opendr.perception.semantic_segmentation.bisenet.algorithm.utils import reverse_one_hot, compute_global_accuracy, \
    fast_hist, per_class_iu
from opendr.engine.learners import Learner
from opendr.engine.data import Image
from opendr.engine.constants import OPENDR_SERVER_URL
from opendr.engine.target import Heatmap


class BisenetLearner(Learner):
    def __init__(self,
                 lr=0.01,
                 iters=1,
                 batch_size=1,
                 optimizer='sgd',
                 temp_path='',
                 checkpoint_after_iter=0,
                 checkpoint_load_iter=0,
                 device='cpu',
                 val_after=1,
                 weight_decay=5e-4,
                 momentum=0.9,
                 drop_last=True,
                 pin_memory=False,
                 num_workers=4,
                 num_classes=12,
                 crop_height=720,
                 crop_width=960,
                 context_path='resnet18'):
        super(BisenetLearner, self).__init__(lr=lr, batch_size=batch_size, iters=iters, optimizer=optimizer,
                                             temp_path=temp_path,
                                             checkpoint_after_iter=checkpoint_after_iter,
                                             checkpoint_load_iter=checkpoint_load_iter, device=device)

        self.val_after = val_after
        self.weight_decay = weight_decay
        self.momentum = momentum
        self.drop_last = drop_last
        self.pin_memory = pin_memory
        self.num_workers = num_workers
        self.num_classes = num_classes
        self.crop_height = crop_height
        self.crop_width = crop_width
        self.context_path = context_path
        self.loss_func = torch.nn.CrossEntropyLoss()
        self.build_model()

        if self.optimizer == "sgd":
            self.optimizer_func = torch.optim.SGD(self.model.parameters(), lr=self.lr, momentum=momentum)
        elif self.optimizer == 'rmsprop':
            self.optimizer_func = torch.optim.RMSprop(self.model.parameters(), lr=self.lr)
        elif self.optimizer == 'adam':
            self.optimizer_func = torch.optim.Adam(self.model.parameters(), lr=self.lr)

    def build_model(self):
        self.model = BiSeNet(num_classes=self.num_classes, context_path=self.context_path)
        if 'cuda' in self.device:
            self.model = torch.nn.DataParallel(self.model).to(self.device)

    def fit(self, dataset, val_dataset=None, silent=False, verbose=True):
        """
        This method is used for training the algorithm on a train dataset

        """

        dataloader_train = DataLoader(dataset,
                                      batch_size=self.batch_size,
                                      shuffle=True,
                                      num_workers=self.num_workers,
                                      pin_memory=self.pin_memory,
                                      drop_last=self.drop_last)

        step = 0
        for epoch in range(1, self.iters + 1):
            self.model.train()
            tq = tqdm.tqdm(total=len(dataloader_train) * self.batch_size)
            tq.set_description('epoch %d, lr %f' % (epoch, self.lr))
            loss_record = []
            for i, (data, label) in enumerate(dataloader_train):
                if 'cuda' in self.device:
                    data = data.to(self.device)
                    label = label.to(self.device)
                output, output_sup1, output_sup2 = self.model(data)
                loss1 = self.loss_func(output, label)
                loss2 = self.loss_func(output_sup1, label)
                loss3 = self.loss_func(output_sup2, label)
                loss = loss1 + loss2 + loss3
                tq.update(self.batch_size)
                tq.set_postfix(loss='%.6f' % loss)
                self.optimizer_func.zero_grad()
                loss.backward()
                self.optimizer_func.step()
                step += 1
                loss_record.append(loss.item())
            tq.close()
            loss_train_mean = np.mean(loss_record)
            if not silent and verbose:
                print('Total Loss : %f' % (loss_train_mean))
            if self.checkpoint_after_iter != 0 and epoch % self.checkpoint_after_iter == 0:
                self.save(os.path.join(self.temp_path, 'epoch', str(epoch)))
            if self.val_after != 0 and epoch % self.val_after == 0 and val_dataset is not None:
                if not silent and verbose:
                    print('Validation...')
                self.eval(val_dataset)

        return {'Total Loss': loss_train_mean}

    def eval(self, dataset, silent=False, verbose=True):

        dataloader_test = DataLoader(dataset,
                                     shuffle=False,
                                     pin_memory=self.pin_memory,
                                     drop_last=False,
                                     num_workers=self.num_workers)

        with torch.no_grad():
            self.model.eval()
            precision_record = []
            tq = tqdm.tqdm(total=len(dataloader_test) * 1)
            tq.set_description('test')
            hist = np.zeros((self.num_classes, self.num_classes))
            for i, (data, label) in enumerate(dataloader_test):
                tq.update(1)
                if 'cuda' in self.device:
                    data = data.to(self.device)
                    label = label.to(self.device)
                predict = self.model(data).squeeze().cpu()
                predict = reverse_one_hot(predict)
                predict = np.array(predict)
                label = label.squeeze().cpu()
                label = np.array(label)
                precision = compute_global_accuracy(predict, label)
                hist += fast_hist(label.flatten(), predict.flatten(), self.num_classes)
                precision_record.append(precision)
            precision = np.mean(precision_record)
            miou_list = per_class_iu(hist)[:-1]
            miou = np.mean(miou_list)
            tq.close()
            if not silent and verbose:
                print('precision for test: %.3f' % precision)
                print('mIoU for validation: %.3f' % miou)

            return {'precision': precision, 'miou': miou}

    def infer(self, img):
        """
        This method is used to perform semantic segmentation on an image.
        It returns a heatmap of the given image.

        """

        if not isinstance(img, Image):
            img = Image(img)
        image = img.convert("channels_last", "rgb")
        resize = iaa.Scale({'height': self.crop_height, 'width': self.crop_width})
        resize_det = resize.to_deterministic()
        image = resize_det.augment_image(image)
        image = PILImage.fromarray(image).convert('RGB')
        image = transforms.ToTensor()(image)
        image = transforms.Normalize((0.485, 0.456, 0.406), (0.229, 0.224, 0.225))(image).unsqueeze(0)
        # predict
        if self.model is None:
            raise UserWarning("No model is loaded, cannot run inference. Load a model first using load().")
        self.model.eval()
        predict = self.model(image).argmax(dim=1).squeeze().cpu()
        heatmap = Heatmap(predict.numpy())
        # optionally save output heatmap as image
        # heatmap_np = heatmap.numpy()
        # colors = np.random.randint(0, 256, (256, 3), dtype=np.uint8)
        # heatmap_o = colors[heatmap_np]
        # heatmap_o = cv2.resize(np.uint8(heatmap_o), (960, 720))
        # sspath = os.path.join(spath, 'heatmap_example.png')
        # cv2.imwrite(sspath, cv2.cvtColor(heatmap_o, cv2.COLOR_RGB2BGR))

        return heatmap

    def download(self, path=None, mode="pretrained", verbose=True,
                 url=OPENDR_SERVER_URL + "perception/semantic_segmentation/bisenet/"):
        """
        Download utility for various Semantic Segmentation components.
        Downloads files depending on mode and
        saves them in the path provided. It supports downloading:
        1)  pretrained model
        2)  testing image

        :param path: Local path to save the files, defaults to self.temp_path if None
        :type path: str, path, optional
        :param mode:  "pretrained", "testingImage", defaults to "pretrained"
        :type mode: str, optional
        :param verbose: Whether to print messages in the console, defaults to False
        :type verbose: bool, optional
        """
        valid_modes = ["pretrained", "testingImage"]

        if mode not in valid_modes:
            raise UserWarning("mode parameter not valid:", mode, ", should be one of:", valid_modes)

        if not os.path.exists(path):
            os.makedirs(path)

        if mode == "pretrained":
            file_url = os.path.join(url, "trainedModels", "bisenet_camvid.pth")

            if verbose:
                print("Downloading pretrained weights...")
            file_path = os.path.join(path, "bisenet_camvid.pth")

            if not os.path.exists(file_path):
                urlretrieve(file_url, file_path)

            file_url = os.path.join(url, "trainedModels", "bisenet_camvid.json")

            if verbose:
                print("Downloading json file...")
            file_path = os.path.join(path, "bisenet_camvid.json")

            if not os.path.exists(file_path):
                urlretrieve(file_url, file_path)

        if mode == "testingImage":
            file_url = os.path.join(url, "datasets", "testImages", "test1.png")
            if verbose:
                print("Downloading a testing image...")
            file_path = os.path.join(path, "test1.png")

            if not os.path.exists(file_path):
                urlretrieve(file_url, file_path)

    def save(self, path, verbose=True):
        if not os.path.isdir(path):
            os.makedirs(path, exist_ok=True)

        model_name = os.path.basename(path)
        model_path = os.path.join(path, model_name + ".pth")
        if verbose:
            print(model_name)
        metadata = {"model_paths": [],
                    "framework": "pytorch",
                    "format": "pth",
                    "has_data": False,
                    "inference_params": {},
                    "optimized": False,
                    "optimizer_info": {}}
        param_filepath = model_name + ".pth"
        metadata["model_paths"].append(param_filepath)
        if self.device == 'cpu':
            torch.save(self.model.state_dict(), model_path)
        elif 'cuda' in self.device:
            torch.save(self.model.module.state_dict(), model_path)
        if verbose:
            print("Model parameters saved.")

        with open(os.path.join(path, model_name + '.json'), 'w', encoding='utf-8') as f:
            json.dump(metadata, f, ensure_ascii=False, indent=4)
        if verbose:
            print("Model metadata saved.")

        return True

    def load(self, path):

        if self.model is None:
            self.build_model()

        if not os.path.isdir(path):
            raise FileNotFoundError(f"Could not find directory {path}")

        folder_basename = os.path.basename(path)
        with open(os.path.join(path, folder_basename + ".json")) as jsonfile:
            metadata = json.load(jsonfile)

        if self.device == 'cpu':
            self.model.load_state_dict(torch.load(os.path.join(path, metadata["model_paths"][0]),
                                                  map_location=torch.device('cpu')),)
        elif 'cuda' in self.device:
            self.model.module.load_state_dict(torch.load(os.path.join(path, metadata["model_paths"][0])))
        self.model.eval()

    def reset(self):
        """This method is not used in this implementation."""
        raise NotImplementedError

    def optimize(self, target_device):
        """This method is not used in this implementation."""
        raise NotImplementedError



import torch.nn as nn
import torch
import numpy as np
import pandas as pd
import random
import numbers
import torchvision


def poly_lr_scheduler(optimizer, init_lr, iter, lr_decay_iter=1,
                      max_iter=300, power=0.9):
    """
    Polynomial decay of learning rate
    :param init_lr is base learning rate
    :param iter is a current iteration
    :param lr_decay_iter how frequently decay occurs, default is 1
    :param power is a polymomial power
    """

    lr = init_lr * (1 - iter / max_iter) ** power
    optimizer.param_groups[0]['lr'] = lr
    return lr


def get_label_info(csv_path):
    # return label -> {label_name: [r_value, g_value, b_value, ...}
    ann = pd.read_csv(csv_path)
    label = {}
    for iter, row in ann.iterrows():
        label_name = row['name']
        r = row['r']
        g = row['g']
        b = row['b']
        class_11 = row['class_11']
        label[label_name] = [int(r), int(g), int(b), class_11]
    return label


def one_hot_it(label, label_info):
    # return semantic_map -> [H, W]
    semantic_map = np.zeros(label.shape[:-1])
    for index, info in enumerate(label_info):
        color = label_info[info]
        equality = np.equal(label, color)
        class_map = np.all(equality, axis=-1)
        semantic_map[class_map] = index

    return semantic_map


def one_hot_it_v11(label, label_info):
    # return semantic_map -> [H, W, class_num]
    semantic_map = np.zeros(label.shape[:-1])
    # from 0 to 11, and 11 means void
    class_index = 0
    for index, info in enumerate(label_info):
        color = label_info[info][:3]
        class_11 = label_info[info][3]
        if class_11 == 1:
            equality = np.equal(label, color)
            class_map = np.all(equality, axis=-1)
            semantic_map[class_map] = class_index
            class_index += 1
        else:
            equality = np.equal(label, color)
            class_map = np.all(equality, axis=-1)
            semantic_map[class_map] = 11
    return semantic_map


def reverse_one_hot(image):
    """
    Transform a 2D array in one-hot format (depth is num_classes),
    to a 2D array with only 1 channel, where each pixel value is
    the classified class key.

    # Arguments
        image: The one-hot format image

    # Returns
        A 2D array with the same width and height as the input, but
        with a depth size of 1, where each pixel value is the classified
        class key.
    """
    image = image.permute(1, 2, 0)
    x = torch.argmax(image, dim=-1)
    return x


def colour_code_segmentation(image, label_values):
    """
    Given a 1-channel array of class keys, colour code the segmentation results.

    # Arguments
        image: single channel array where each value represents the class key.
        label_values

    # Returns
        Colour coded image for segmentation visualization
    """

    label_values = [label_values[key][:3] for key in label_values if label_values[key][3] == 1]
    label_values.append([0, 0, 0])
    colour_codes = np.array(label_values)
    x = colour_codes[image.astype(int)]

    return x


def compute_global_accuracy(pred, label):
    pred = pred.flatten()
    label = label.flatten()
    total = len(label)
    count = 0.0
    for i in range(total):
        if pred[i] == label[i]:
            count = count + 1.0
    return float(count) / float(total)


def fast_hist(a, b, n):
    '''
    a and b are predict and mask respectively
    n is the number of classes
    '''
    k = (a >= 0) & (a < n)
    return np.bincount(n * a[k].astype(int) + b[k], minlength=n ** 2).reshape(n, n)


def per_class_iu(hist):
    epsilon = 1e-5
    return (np.diag(hist) + epsilon) / (hist.sum(1) + hist.sum(0) - np.diag(hist) + epsilon)


class RandomCrop(object):
    """Crop the given PIL Image at a random location.

    Args:
        size (sequence or int): Desired output size of the crop. If size is an
            int instead of sequence like (h, w), a square crop (size, size) is
            made.
        padding (int or sequence, optional): Optional padding on each border
            of the image. Default is 0, i.e no padding. If a sequence of length
            4 is provided, it is used to pad left, top, right, bottom borders
            respectively.
        pad_if_needed (boolean): It will pad the image if smaller than the
            desired size to avoid raising an exception.
    """

    def __init__(self, size, seed, padding=0, pad_if_needed=False):
        if isinstance(size, numbers.Number):
            self.size = (int(size), int(size))
        else:
            self.size = size
        self.padding = padding
        self.pad_if_needed = pad_if_needed
        self.seed = seed

    @staticmethod
    def get_params(img, output_size, seed):
        """Get parameters for ``crop`` for a random crop.

        Args:
            img (PIL Image): Image to be cropped.
            output_size (tuple): Expected output size of the crop.

        Returns:
            tuple: params (i, j, h, w) to be passed to ``crop`` for random crop.
        """
        random.seed(seed)
        w, h = img.size
        th, tw = output_size
        if w == tw and h == th:
            return 0, 0, h, w
        i = random.randint(0, h - th)
        j = random.randint(0, w - tw)
        return i, j, th, tw

    def __call__(self, img):
        """
        Args:
            img (PIL Image): Image to be cropped.

        Returns:
            PIL Image: Cropped image.
        """
        if self.padding > 0:
            img = torchvision.transforms.functional.pad(img, self.padding)

        # pad the width if needed
        if self.pad_if_needed and img.size[0] < self.size[1]:
            img = torchvision.transforms.functional.pad(img, (int((1 + self.size[1] - img.size[0]) / 2), 0))
        # pad the height if needed
        if self.pad_if_needed and img.size[1] < self.size[0]:
            img = torchvision.transforms.functional.pad(img, (0, int((1 + self.size[0] - img.size[1]) / 2)))

        i, j, h, w = self.get_params(img, self.size, self.seed)

        return torchvision.transforms.functional.crop(img, i, j, h, w)

    def __repr__(self):
        return self.__class__.__name__ + '(size={0}, padding={1})'.format(self.size, self.padding)


def cal_miou(miou_list, csv_path):
    # return label -> {label_name: [r_value, g_value, b_value, ...}
    ann = pd.read_csv(csv_path)
    miou_dict = {}
    cnt = 0
    for iter, row in ann.iterrows():
        label_name = row['name']
        class_11 = int(row['class_11'])
        if class_11 == 1:
            miou_dict[label_name] = miou_list[cnt]
            cnt += 1
    return miou_dict, np.mean(miou_list)


class OHEM_CrossEntroy_Loss(nn.Module):
    def __init__(self, threshold, keep_num):
        super(OHEM_CrossEntroy_Loss, self).__init__()
        self.threshold = threshold
        self.keep_num = keep_num
        self.loss_function = nn.CrossEntropyLoss(reduction='none')

    def forward(self, output, target):
        loss = self.loss_function(output, target).view(-1)
        loss, loss_index = torch.sort(loss, descending=True)
        threshold_in_keep_num = loss[self.keep_num]
        if threshold_in_keep_num > self.threshold:
            loss = loss[loss > self.threshold]
        else:
            loss = loss[:self.keep_num]
        return torch.mean(loss)


def group_weight(weight_group, module, norm_layer, lr):
    group_decay = []
    group_no_decay = []
    for m in module.modules():
        if isinstance(m, nn.Linear):
            group_decay.append(m.weight)
            if m.bias is not None:
                group_no_decay.append(m.bias)
        elif isinstance(m, (nn.Conv2d, nn.Conv3d)):
            group_decay.append(m.weight)
            if m.bias is not None:
                group_no_decay.append(m.bias)
        elif isinstance(m, norm_layer) or isinstance(m, nn.GroupNorm):
            if m.weight is not None:
                group_no_decay.append(m.weight)
            if m.bias is not None:
                group_no_decay.append(m.bias)

    assert len(list(module.parameters())) == len(group_decay) + len(
        group_no_decay)
    weight_group.append(dict(params=group_decay, lr=lr))
    weight_group.append(dict(params=group_no_decay, weight_decay=.0, lr=lr))
    return weight_group



import torch
from torchvision import models


class resnet18(torch.nn.Module):
    def __init__(self, pretrained=True):
        super().__init__()
        self.features = models.resnet18(pretrained=pretrained)
        self.conv1 = self.features.conv1
        self.bn1 = self.features.bn1
        self.relu = self.features.relu
        self.maxpool1 = self.features.maxpool
        self.layer1 = self.features.layer1
        self.layer2 = self.features.layer2
        self.layer3 = self.features.layer3
        self.layer4 = self.features.layer4

    def forward(self, input):
        x = self.conv1(input)
        x = self.relu(self.bn1(x))
        x = self.maxpool1(x)
        feature1 = self.layer1(x)  # 1 / 4
        feature2 = self.layer2(feature1)  # 1 / 8
        feature3 = self.layer3(feature2)  # 1 / 16
        feature4 = self.layer4(feature3)  # 1 / 32
        # global average pooling to build tail
        tail = torch.mean(feature4, 3, keepdim=True)
        tail = torch.mean(tail, 2, keepdim=True)
        return feature3, feature4, tail


class resnet101(torch.nn.Module):
    def __init__(self, pretrained=True):
        super().__init__()
        self.features = models.resnet101(pretrained=pretrained)
        self.conv1 = self.features.conv1
        self.bn1 = self.features.bn1
        self.relu = self.features.relu
        self.maxpool1 = self.features.maxpool
        self.layer1 = self.features.layer1
        self.layer2 = self.features.layer2
        self.layer3 = self.features.layer3
        self.layer4 = self.features.layer4

    def forward(self, input):
        x = self.conv1(input)
        x = self.relu(self.bn1(x))
        x = self.maxpool1(x)
        feature1 = self.layer1(x)  # 1 / 4
        feature2 = self.layer2(feature1)  # 1 / 8
        feature3 = self.layer3(feature2)  # 1 / 16
        feature4 = self.layer4(feature3)  # 1 / 32
        # global average pooling to build tail
        tail = torch.mean(feature4, 3, keepdim=True)
        tail = torch.mean(tail, 2, keepdim=True)
        return feature3, feature4, tail


def build_contextpath(name):
    model = {
        'resnet18': resnet18(pretrained=True),
        'resnet101': resnet101(pretrained=True)
    }
    return model[name]


if __name__ == '__main__':
    #
    model_18 = build_contextpath('resnet18')
    model_101 = build_contextpath('resnet101')
    x = torch.rand(1, 3, 256, 256)

    y_18 = model_18(x)
    y_101 = model_101(x)



import torch
from torch import nn
from opendr.perception.semantic_segmentation.bisenet.algorithm.model.build_contextpath import build_contextpath
import warnings

warnings.filterwarnings(action='ignore')


class ConvBlock(torch.nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size=3, stride=2, padding=1):
        super().__init__()
        self.conv1 = nn.Conv2d(in_channels, out_channels, kernel_size=kernel_size, stride=stride, padding=padding,
                               bias=False)
        self.bn = nn.BatchNorm2d(out_channels)
        self.relu = nn.ReLU()

    def forward(self, input):
        x = self.conv1(input)
        return self.relu(self.bn(x))


class Spatial_path(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.convblock1 = ConvBlock(in_channels=3, out_channels=64)
        self.convblock2 = ConvBlock(in_channels=64, out_channels=128)
        self.convblock3 = ConvBlock(in_channels=128, out_channels=256)

    def forward(self, input):
        x = self.convblock1(input)
        x = self.convblock2(x)
        x = self.convblock3(x)
        return x


class AttentionRefinementModule(torch.nn.Module):
    def __init__(self, in_channels, out_channels):
        super().__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size=1)
        self.bn = nn.BatchNorm2d(out_channels)
        self.sigmoid = nn.Sigmoid()
        self.in_channels = in_channels
        self.avgpool = nn.AdaptiveAvgPool2d(output_size=(1, 1))

    def forward(self, input):
        # global average pooling
        x = self.avgpool(input)
        assert self.in_channels == x.size(1), 'in_channels and out_channels should all be {}'.format(x.size(1))
        x = self.conv(x)
        x = self.sigmoid(x)
        # channels of input and x should be same
        x = torch.mul(input, x)
        return x


class FeatureFusionModule(torch.nn.Module):
    def __init__(self, num_classes, in_channels):
        super().__init__()
        self.in_channels = in_channels

        self.convblock = ConvBlock(in_channels=self.in_channels, out_channels=num_classes, stride=1)
        self.conv1 = nn.Conv2d(num_classes, num_classes, kernel_size=1)
        self.relu = nn.ReLU()
        self.conv2 = nn.Conv2d(num_classes, num_classes, kernel_size=1)
        self.sigmoid = nn.Sigmoid()
        self.avgpool = nn.AdaptiveAvgPool2d(output_size=(1, 1))

    def forward(self, input_1, input_2):
        x = torch.cat((input_1, input_2), dim=1)
        assert self.in_channels == x.size(1), 'in_channels of ConvBlock should be {}'.format(x.size(1))
        feature = self.convblock(x)
        x = self.avgpool(feature)

        x = self.relu(self.conv1(x))
        x = self.sigmoid(self.conv2(x))
        x = torch.mul(feature, x)
        x = torch.add(x, feature)
        return x


class BiSeNet(torch.nn.Module):
    def __init__(self, num_classes, context_path):
        super().__init__()
        # build spatial path
        self.saptial_path = Spatial_path()

        # build context path
        self.context_path = build_contextpath(name=context_path)

        # build attention refinement module for resnet 101
        if context_path == 'resnet101':
            self.attention_refinement_module1 = AttentionRefinementModule(1024, 1024)
            self.attention_refinement_module2 = AttentionRefinementModule(2048, 2048)
            # supervision block
            self.supervision1 = nn.Conv2d(in_channels=1024, out_channels=num_classes, kernel_size=1)
            self.supervision2 = nn.Conv2d(in_channels=2048, out_channels=num_classes, kernel_size=1)
            # build feature fusion module
            self.feature_fusion_module = FeatureFusionModule(num_classes, 3328)

        elif context_path == 'resnet18':
            # build attention refinement module  for resnet 18
            self.attention_refinement_module1 = AttentionRefinementModule(256, 256)
            self.attention_refinement_module2 = AttentionRefinementModule(512, 512)
            # supervision block
            self.supervision1 = nn.Conv2d(in_channels=256, out_channels=num_classes, kernel_size=1)
            self.supervision2 = nn.Conv2d(in_channels=512, out_channels=num_classes, kernel_size=1)
            # build feature fusion module
            self.feature_fusion_module = FeatureFusionModule(num_classes, 1024)
        else:
            raise ValueError("unsupport context_path network")

        # build final convolution
        self.conv = nn.Conv2d(in_channels=num_classes, out_channels=num_classes, kernel_size=1)

        self.init_weight()

        self.mul_lr = []
        self.mul_lr.append(self.saptial_path)
        self.mul_lr.append(self.attention_refinement_module1)
        self.mul_lr.append(self.attention_refinement_module2)
        self.mul_lr.append(self.supervision1)
        self.mul_lr.append(self.supervision2)
        self.mul_lr.append(self.feature_fusion_module)
        self.mul_lr.append(self.conv)

    def init_weight(self):
        for name, m in self.named_modules():
            if 'context_path' not in name:
                if isinstance(m, nn.Conv2d):
                    nn.init.kaiming_normal_(m.weight, mode='fan_in', nonlinearity='relu')
                elif isinstance(m, nn.BatchNorm2d):
                    m.eps = 1e-5
                    m.momentum = 0.1
                    nn.init.constant_(m.weight, 1)
                    nn.init.constant_(m.bias, 0)

    def forward(self, input):
        # output of spatial path
        sx = self.saptial_path(input)

        # output of context path
        cx1, cx2, tail = self.context_path(input)
        cx1 = self.attention_refinement_module1(cx1)
        cx2 = self.attention_refinement_module2(cx2)
        cx2 = torch.mul(cx2, tail)
        # upsampling
        cx1 = torch.nn.functional.interpolate(cx1, size=sx.size()[-2:], mode='bilinear')
        cx2 = torch.nn.functional.interpolate(cx2, size=sx.size()[-2:], mode='bilinear')
        cx = torch.cat((cx1, cx2), dim=1)

        if self.training:
            cx1_sup = self.supervision1(cx1)
            cx2_sup = self.supervision2(cx2)
            cx1_sup = torch.nn.functional.interpolate(cx1_sup, size=input.size()[-2:], mode='bilinear')
            cx2_sup = torch.nn.functional.interpolate(cx2_sup, size=input.size()[-2:], mode='bilinear')

        # output of feature fusion module
        result = self.feature_fusion_module(sx, cx)

        # upsampling
        result = torch.nn.functional.interpolate(result, scale_factor=8, mode='bilinear')
        result = self.conv(result)

        if self.training:
            return result, cx1_sup, cx2_sup

        return result


if __name__ == '__main__':
    import os

    os.environ['CUDA_VISIBLE_DEVICES'] = '0'
    model = BiSeNet(32, 'resnet18')

    model = model.cuda()
    x = torch.rand(2, 3, 256, 256)
    record = model.parameters()

    print(model.parameters())



# Copyright 2020-2022 OpenDR European Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import glob
import os
from torchvision import transforms
from PIL import Image
import numpy as np
from imgaug import augmenters as iaa
import random
import zipfile
from urllib.request import urlretrieve
from opendr.perception.semantic_segmentation.bisenet.algorithm.utils import get_label_info, RandomCrop, one_hot_it_v11
from opendr.engine.datasets import ExternalDataset, DatasetIterator
from opendr.engine.constants import OPENDR_SERVER_URL


class CamVidDataset(ExternalDataset, DatasetIterator, torch.utils.data.Dataset):
    """
    CamVid dataset for Semantic Segmentation
    path (dir): path directory of the CamVid dataset, containinig the subfolders "train",
     "train_labels", "test", "test_labels", and the class_dict.csv file
    scale (int): (crop_height, crop_width)
    mode: 'test', 'train'
    """

    def __init__(self,
                 path: str,
                 scale=(720, 960),
                 mode='train'):
        ExternalDataset.__init__(self, path=str(path), dataset_type="camvid")
        DatasetIterator.__init__(self)
        torch.utils.data.Dataset.__init__(self)

        self.mode = mode
        if self.mode == 'train':
            self.image_path = os.path.join(self.path, 'train')
            self.label_path = os.path.join(self.path, 'train_labels')
        elif self.mode == 'test':
            self.image_path = os.path.join(self.path, 'test')
            self.label_path = os.path.join(self.path, 'test_labels')
        self.image_list = []
        if not isinstance(self.image_path, list):
            self.image_path = [self.image_path]
        for image_path_ in self.image_path:
            self.image_list.extend(glob.glob(os.path.join(image_path_, '*.png')))
        self.image_list.sort()
        self.label_list = []
        if not isinstance(self.label_path, list):
            self.label_path = [self.label_path]
        for label_path_ in self.label_path:
            self.label_list.extend(glob.glob(os.path.join(label_path_, '*.png')))
        self.label_list.sort()
        self.fliplr = iaa.Fliplr(0.5)
        self.label_info = get_label_info(os.path.join(path, "class_dict.csv"))
        self.to_tensor = transforms.Compose([
            transforms.ToTensor(),
            transforms.Normalize((0.485, 0.456, 0.406), (0.229, 0.224, 0.225)),
        ])
        self.image_size = scale
        self.scale = [0.5, 1, 1.25, 1.5, 1.75, 2]

    def __getitem__(self, index):
        # load image and crop
        seed = random.random()
        img = Image.open(self.image_list[index])
        scale = random.choice(self.scale)
        scale = (int(self.image_size[0] * scale), int(self.image_size[1] * scale))
        if self.mode == 'train':
            img = transforms.Resize(scale, Image.BILINEAR)(img)
            img = RandomCrop(self.image_size, seed, pad_if_needed=True)(img)
        # =====================================
        img = np.array(img)
        # load label
        label = Image.open(self.label_list[index])
        if self.mode == 'train':
            label = transforms.Resize(scale, Image.NEAREST)(label)
            label = RandomCrop(self.image_size, seed, pad_if_needed=True)(label)
        # =====================================
        label = np.array(label)
        # augment image and label
        if self.mode == 'train':
            seq_det = self.fliplr.to_deterministic()
            img = seq_det.augment_image(img)
            label = seq_det.augment_image(label)
        # image -> [C, H, W]
        img = Image.fromarray(img)
        img = self.to_tensor(img).float()
        label = one_hot_it_v11(label, self.label_info).astype(np.uint8)
        label = torch.from_numpy(label).long()

        return img, label

    def __len__(self):
        return len(self.image_list)

    @staticmethod
    def download_data(path: str):
        """Download camvid dataset

        Args:
            path: Directory in which to store the dataset
        """
        os.makedirs(path)
        url = os.path.join(
            OPENDR_SERVER_URL,
            "perception",
            "semantic_segmentation",
            "bisenet",
            "datasets",
            "CamVid.zip"
        )
        zip_path = os.path.join(path, 'CamVid.zip')
        urlretrieve(url=url, filename=zip_path)
        with zipfile.ZipFile(zip_path, 'r') as zip_ref:
            zip_ref.extractall(path)
        os.remove(zip_path)
